"""
LavaVu python interface

Interactive HTML UI controls library
These functions can be called from the ControlFactory provided on

- lavavu.Viewer()
- lavavu.Object()
- lavavu.Properties()

Some work only on the viewer and some only to objects and some to both.
Each will provide a control that allows modifying visualisation properties or executing commands interactively
in an IPython notebook environment

Example
-------

Create a Button control to execute a rotation command:

>>> lv = lavavu.Viewer()
>>> lv.control.Button('rotate x 1')

Create a Range control to adjust the pointsize property:

>>> lv = lavavu.Viewer()
>>> pts = lv.points()
>>> pts.control.Range('pointsize', range=[1, 10])

For convenience, controls to set visualisation properties can often be created by calling control directly with the property name
This will create a control of the default type for for that property if one is defined, e.g.

>>> lv = lavavu.Viewer()
>>> pts = lv.points()
>>> pts.control('pointsize')
"""

__all__ = ['Button', 'Checkbox', 'Colour', 'ColourMapList', 'ColourMaps', 'Command', 'Divider', 'DualRange', 'Entry', 'File', 'Filter', 'Gradient', 'List', 'Number', 'Number2D', 'Number3D', 'ObjectList', 'ObjectSelect', 'Panel', 'Range', 'Range2D', 'Range3D', 'Rotation', 'Tabs', 'TimeStepper', 'Window']

import os
import sys
import time
import datetime
import json
from vutils import is_ipython, is_notebook
import weakref
import string
import re
from random import Random

#Register of windows (viewer instances)
windows = []
#Window unique ids
winids = []
#Generate unique strings for controls and windows
id_random = Random() #Ensure we use our own default seed in case set in notebook
def gen_id(length=10):
    alphabet = string.ascii_letters + string.digits
    try:
        return ''.join(id_random.choices(alphabet, k=length))
    except:
        #Python 3.5
        return ''.join(id_random.choice(string.ascii_uppercase + string.ascii_lowercase + string.digits) for _ in range(length))

basehtml = """
<html>

<head>
<title>LavaVu Interface</title>
<meta http-equiv="content-type" content="text/html; charset=ISO-8859-1">

<style>
html, body, table, form {
  padding: 0; margin: 0;
  border: none;
  font-size: 9pt;
  font-family: sans-serif;
}
p {margin: 0px 2px; }

.canvas {
  width: 100%; height: 100%;
}

canvas {
  z-index: 0; margin: 0px; padding:0px;
  border: none;
}

</style>

---SCRIPTS---

</head>

<body onload="---INIT---" oncontextmenu="return false;">
---CONTENT---
</body>

</html>
"""

inlinehtml = """
---SCRIPTS---
<div id="---ID---" style="display: block; width: ---WIDTH---px; height: ---HEIGHT---px;"></div>
---CONTENT---

<script>
---INIT---
</script>
"""

#Some common elements (TODO: dynamically create these when not found)
hiddenhtml = """
<div id="progress" class="popup" style="display: none; width: 310px; height: 32px;">
  <span id="progressmessage"></span><span id="progressstatus"></span>
  <div id="progressbar" style="width: 300px; height: 10px; background: #58f;"></div>
</div>

<div id="hidden" style="display: none">
  <img src="data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAkAAAAPCAYAAAA2yOUNAAAAj0lEQVQokWNIjHT8/+zZs//Pnj37/+TJk/9XLp/+f+bEwf9HDm79v2Prqv9aKrz/GUYVEaeoMDMQryJXayWIoi0bFmFV1NWS+z/E1/Q/AwMDA0NVcez/LRsWoSia2luOUAADVcWx/xfO6/1/5fLp/1N7y//HhlmhKoCBgoyA/w3Vyf8jgyyxK4CBUF8zDAUAAJRXY0G1eRgAAAAASUVORK5CYII=" id="slider">
  <canvas id="gradient" width="2048" height="1"></canvas>
  <canvas id="palette" width="512" height="24" class="palette checkerboard"></canvas>
</div>
"""

#Global data (script) - set from lavavu
jsglobals = ''

#Static HTML location
htmlpath = ""

def _isviewer(target):
    """Return true if target is a viewer"""
    #return not hasattr(target, "parent")
    return target.__class__.__name__ == 'Viewer'

def _isobject(target):
    """Return true if target is a lavavu object"""
    return target.__class__.__name__ == 'Object'

def _getviewer(target):
    """Return its viewer if target is vis object
    otherwise just return target as if is a viewer"""
    if not _isviewer(target):
        return target.parent
    return target

def _getproperty(target, propname):
    """Return value of property from target
    if not defined, return the default value for this property
    """
    _lv = _getviewer(target)
    if propname in target:
        return target[propname]
    elif propname in _lv.properties:
        #Get property default
        prop = _lv.properties[propname]
        return prop["default"]
    else:
        return None

_file_cache = dict()

def _webglcode(shaders, css, scripts, menu=True, lighttheme=True, stats=False):
    """
    Returns base WebGL code, by default using full WebGL renderer (draw.js)
    Pass 'drawbox.js' for box interactor only
    """
    code = shaders
    if menu:
        scripts[1] += ['menu.js']
        #Following scripts need require disabled
        scripts[0] += ['dat.gui.min.js']
        #Stats module for WebVR
        if stats:
            scripts[0] += ['stats.min.js']
        if lighttheme:
            css.append("dat-gui-light-theme.css")
        #Some css tweaks
        css.append("gui.css")

    else:
        #DAT.gui disabled
        scripts[0] += ['!GUI_OFF']

    #First scripts list needs require.js disabled
    code += _getjslibs(['!REQUIRE_OFF'] + scripts[0] + ['!REQUIRE_ON'] + scripts[1])

    code += _getcss(css)
    code += '<script>\n' + jsglobals + '\n</script>\n'
    return code

def _webglviewcode(shaderpath, menu=True, lighttheme=True):
    """
    Returns WebGL base code for an interactive visualisation window
    """
    jslibs = [['gl-matrix-min.js'], ['OK-min.js', 'draw.js']]
    return _webglcode(_getshaders(shaderpath), ['styles.css'], jslibs, menu=menu, lighttheme=lighttheme)

def _webglboxcode(menu=True, lighttheme=True):
    """
    Returns WebGL base code for the bounding box interactor window
    """
    jslibs = [['gl-matrix-min.js'], ['OK-min.js', 'control.js', 'drawbox.js']]
    return _webglcode('', ['control.css', 'styles.css'], jslibs, menu=menu, lighttheme=lighttheme)

def _connectcode(target):
    """
    Returns base code to open a connection only, for obtaining correct connection url
    (Only required if no interaction windows have been opened yet)
    """
    #If connection via window already opened, we don't need this
    if len(winids) == 0:
        initjs = '\n_wi[0] = new WindowInteractor(0, {uid}, {port});\n;'.format(uid=id(target), port=target.port)
        #return _readfilehtml('control.js') + initjs
        s = '<script>\n' + _readfilehtml('control.js') + initjs + '</script>\n'
        return s
    else:
        return ""

def _getcss(files=["styles.css"]):
    #Load stylesheets to inline tag
    return _filestohtml(files, tag="style")

def _getjslibs(files):
    #Load a set of combined javascript libraries in script tags
    return _filestohtml(files, tag="script")

def _filestohtml(files, tag="script"):
    #Load a set of files into a string enclosed in specified html tag
    code = '<' + tag + '>\n'
    for f in files:
        code += _readfilehtml(f)
    code += '\n</' + tag + '>\n'
    return code

def _readfilehtml(filename):
    #HACK: Need to disable require.js to load libraries from inline script tags
    if filename == "!REQUIRE_OFF":
        return '_backup_define = window.define; window.define = undefined;'
    elif filename == "!REQUIRE_ON":
        return 'window.define = _backup_define; delete _backup_define;'
    elif filename == "!GUI_OFF":
        return 'window.dat = undefined;'

    #Read a file from the htmlpath (or cached copy)
    global _file_cache
    if not filename in _file_cache:
        _file_cache[filename] = _readfile(os.path.join(htmlpath, filename))
    return _file_cache[filename]

def _readfile(filename):
    #Read a text file and return contents
    data = ""
    with open(filename, 'r') as f:
        data = f.read()
        f.close()
    return data

def _getshaders(path, shaders=['points', 'lines', 'triangles', 'volume']):
    #Load combined shaders
    src = '<script>\nvar shaders = {};\n'
    sdict = {'points' : 'point', 'lines' : 'line', 'triangles' : 'tri', 'volume' : 'volume'};
    for key in shaders:
        vs = _readfile(os.path.join(path, sdict[key] + 'Shader.vert'))
        fs = _readfile(os.path.join(path, sdict[key] + 'Shader.frag'))
        vs = vs.replace('"', '\\"').replace('\n', '\\n')
        fs = fs.replace('"', '\\"').replace('\n', '\\n')
        #vs = re.sub("^\s*in ", 'attribute ', vs, flags=re.MULTILINE);
        #vs = re.sub("^\s*out ", 'varying ', vs, flags=re.MULTILINE)
        #fs = re.sub("^\s*in ", 'varying ', fs, flags=re.MULTILINE);
        src += 'shaders["' + key + '-vs"] = "' + vs + '";'
        src += 'shaders["' + key + '-fs"] = "' + fs + '";'
    src += '\n</script>\n\n'
    return src

def _getshaders_as_scripts(path, shaders=['points', 'lines', 'triangles', 'volume']):
    #Load shaders into script tags, not compatible with colab for some reason
    src = ''
    sdict = {'points' : 'point', 'lines' : 'line', 'triangles' : 'tri', 'volume' : 'volume'};
    for key in shaders:
        src += '<script id="' + key + '-vs" type="x-shader/x-vertex">\n'
        src += _readfile(os.path.join(path, sdict[key] + 'Shader.vert'))
        src += '</script>\n\n'
        src += '<script id="' + key + '-fs" type="x-shader/x-fragment">\n'
        src += _readfile(os.path.join(path, sdict[key] + 'Shader.frag'))
        src += '</script>\n\n'
    return src

class _Action(object):
    """Base class for an action triggered by a control

    Default action is to run a command
    """

    def __init__(self, target, command=None, readproperty=None):
        self.target = target
        self.command = command
        if not hasattr(self, "property"):
            self.property = readproperty
        self.lastvalue = 0
        self.index = None

    def script(self):
        #Return script action for HTML export
        if self.command is None or not len(self.command): return ""
        #Run a command with a value argument
        return self.command + ' ---VAL---'

class _PropertyAction(_Action):
    """Property change action triggered by a control
    """

    def __init__(self, target, prop, command=None, index=None):
        self.property = prop
        #Default action after property set is redraw, can be set to provided
        #if command is None:
        #    command = "redraw"
        self.command = command
        super(_PropertyAction, self).__init__(target, command)
        self.index = index

    def script(self):
        #Return script action for HTML export
        #Set a property
        #Check for index (3D prop)
        propset = self.property + '=---VAL---'
        if self.index is not None:
            propset = self.property + '[' + str(self.index) + ']=---VAL---'
        # - Globally
        script = ''
        if _isviewer(self.target):
            script = 'select; ' + propset
        # - on an object selector, select the object
        elif type(self.target).__name__ == 'ObjectSelect':
            script = propset
        # - on a properties collection
        elif type(self.target).__name__ == 'Properties':
            script = '$' + str(id(self.target)) + ' ' + self.property + ' ---VAL--- '
        # - On an object
        else:
            script = '<' + self.target["name"] + '>' + propset
        #Add any additional commands
        return script + '; ' + super(_PropertyAction, self).script()


class _CommandAction(_Action):
    """Command action triggered by a control, with object select before command
    """

    def __init__(self, target, command, readproperty):
        self.command = command
        super(_CommandAction, self).__init__(target, command, readproperty)

    def script(self):
        #Return script action for HTML export
        #Set a property
        # - Globally
        script = ''
        if _isviewer(self.target):
            script = 'select; '
        # - on an object selector, select the object
        elif type(self.target).__name__ == 'ObjectSelect':
            script = ''
        # - On an object
        else:
            script = '<' + self.target["name"] + '>'
        #Add the commands
        return script + super(_CommandAction, self).script()

class _FilterAction(_PropertyAction):
    """Filter property change action triggered by a control
    """
    def __init__(self, target, findex, prop, command=None):
        self.findex = findex
        if command is None: command = "redraw"
        self.command = command
        super(_FilterAction, self).__init__(target, prop)

    def script(self):
        #Return script action for HTML export
        #Set a filter range
        cmd = "filtermin" if self.property == "minimum" else "filtermax"
        return 'select ' + self.target["name"] + '; ' + cmd + ' ' + str(self.findex) + ' ---VAL---'
        #propset = "filters=" + json.dumps()
        #script = 'select ' + self.target["name"] + '; ' + propset

class _HTML(object):
    """A class to output HTML controls
    """

    #Parent class for container types
    def __init__(self, label):
        self.label = label
        self.uniqueid()

    def html(self):
        """Return the HTML code"""
        return ''

    def uniqueid(self):
        #Get a unique control identifier
        self.id = gen_id()
        self.elid = "lvctrl_" + self.id

    def labelhtml(self):
        #Default label
        html = ""
        if self.label:
            html += '<p>' + self.label + ':</p>\n'
        return html


class _Container(_HTML):
    """A container for a set of controls
    """
    #Parent class for container types
    def __init__(self, viewer, label=""):
        self.viewer = viewer
        self._content = []
        self._current = 0
        super(_Container, self).__init__(label)

    def add(self, ctrl):
        self._content.append(ctrl)
        return ctrl

    def controls(self):
        return self.html()

    def __iter__(self):
        return self

    def __next__(self):
        if self._current > len(self._content)-1:
            self._current = 0
            raise StopIteration
        else:
            self._current += 1
            return self._content[self._current-1]

    def html(self):
        html = self.labelhtml()
        html += '<div style="padding:0px; margin: 0px; position: relative;" class="lvctrl">\n'
        for i in range(len(self._content)):
            html += self._content[i].controls()
        html += '</div>\n'
        return html

    def scripts(self):
        #Returns script dictionary
        d = {}
        for i in range(len(self._content)):
            d.update(self._content[i].scripts())
        return d

class Window(_Container):
    """
    Creates an interaction window with an image of the viewer frame 
    and webgl controller for rotation/translation

    Parameters
    ----------
    align : str
        Set to "left/right" to align viewer window, default is left. Set to None to skip.
    wrapper : str
        Set the style of the wrapper div, default is empty string so wrapper is enabled with no custom style
        Set to None to disable wrapper
    """
    def __init__(self, viewer, resolution=None, align="left", wrapper=""):
        super(Window, self).__init__(viewer)
        self.align = align
        self.wrapper = wrapper
        if resolution is not None:
            viewer.output_resolution = resolution

    def html(self):
        style = 'min-height: 50px; min-width: 50px; position: relative; display: inline-block; '
        if self.align is not None:
            style += 'float: ' + self.align + ';'
        if self.wrapper is not None:
            style += ' margin-right: 10px;'
        html = ""
        html += '<div style="' + style + '">\n'
        html += '<img id="imgtarget_---VIEWERID---" draggable=false style="margin: 0px; border: 1px solid #aaa; display: inline-block;" src="data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAEAAAABACAQAAAAAYLlVAAAAPUlEQVR42u3OMQEAAAgDINe/iSU1xh5IQPamKgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgLtwAMBsGqBDct9xQAAAABJRU5ErkJggg==">\n'
        html += """
           <div style="display: none; z-index: 200; position: absolute; top: 5px; right: 5px;">
             <select onchange="_wi['---VIEWERID---'].box.mode = this.value;">
               <option>Rotate</option>
               <option>Translate</option>
               <option>Zoom</option>
             </select>
             <input type="button" value="Reset" onclick="_wi['---VIEWERID---'].execute('reset');">
           </div>"""
        html += '</div>\n'

        #Display any contained controls
        if self.wrapper is not None:
            html += '<div style="' + self.wrapper + '" class="lvctrl">\n'
        html += super(Window, self).html()
        if self.wrapper is not None:
            html += '</div>\n'
        #html += '<div style="clear: both;">\n'
        return html

class FillWindow(_Container):
    """
    Creates an interaction window with an image of the viewer frame 
    and webgl controller for rotation/translation

    Unlike Window, which uses the size of the returned image to set the interaction canvas size,
    FillWindow creates an interaction canvas that fills the parent element width completely.
    Images from the renderer are requested to match that size without rescaling.

    Parameters
    ----------
    aspect_ratio : tuple (int,int)
        Aspect ratio to calculate the height, eg: (16,9) (default) or (4,3)
    minwidth : int
        Minimum width of the image
    minheight : int
        Minimum height of the image
    """
    def __init__(self, viewer, aspect_ratio=(16,9), minwidth=100, minheight=50):
        super(FillWindow, self).__init__(viewer)
        if aspect_ratio is not None and len(aspect_ratio) == 2:
            viewer.output_resolution = (viewer.output_resolution[0], int(viewer.output_resolution[0] * aspect_ratio[1] / aspect_ratio[0]))
        self.minwidth = minwidth
        self.minheight = minheight

    def html(self):
        style = 'height: 100%; width: 100%; position: relative; display: inline-block; '
        style += 'min-width: ' + str(self.minwidth) + 'px; '
        style += 'min-height: ' + str(self.minheight) + 'px; '
        html = ""
        html += '<div>\n'
        html += '<img id="imgtarget_---VIEWERID---" draggable=false style="' + style + 'margin: 0px; border: 0px; display: inline-block;" src="data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAEAAAABACAQAAAAAYLlVAAAAPUlEQVR42u3OMQEAAAgDINe/iSU1xh5IQPamKgICAgICAgICAgICAgICAgICAgICAgICAgICAgICAgLtwAMBsGqBDct9xQAAAABJRU5ErkJggg==">\n'
        html += """
           <div style="display: none; z-index: 200; position: absolute; top: 5px; right: 5px;">
             <select onchange="_wi['---VIEWERID---'].box.mode = this.value;">
               <option>Rotate</option>
               <option>Translate</option>
               <option>Zoom</option>
             </select>
             <input type="button" value="Reset" onclick="_wi['---VIEWERID---'].execute('reset');">
           </div>"""
        html += '</div>\n'

        #Display any contained controls
        html += super(FillWindow, self).html()
        return html

class Panel(Window):
    """Creates a control panel with an interactive viewer window and a set of controls
    placed to the left with the viewer aligned to the right
    """
    def __init__(self, *args, **kwargs):
        super(Panel, self).__init__(*args, align="right", wrapper=None, **kwargs)

class Tabs(_Container):
    """Creates a group of controls with tabs that can be shown or hidden

    Parameters
    ----------
    buttons : boolean
        Display the tab buttons for switching tabs
    """
    def __init__(self, target, buttons=True):
        self.tabs = []
        self.buttons = buttons
        super(Tabs, self).__init__(target)

    def tab(self, label=""):
        """Add a new tab, any controls appending will appear in the new tab
        Parameters
        ----------
        label : str
            Label for the tab, if omitted will be blank
        """
        self.tabs.append(label)

    def add(self, ctrl):
        if not len(self.tabs): self.tab()
        self._content.append(ctrl)
        ctrl.tab = len(self.tabs)-1

    def html(self):
        html = """
        <script>
        function openTab_---ELID---(el, tabName) {
          var i;
          var x = document.getElementsByClassName("---ELID---");
          for (i = 0; i < x.length; i++)
             x[i].style.display = "none";  
          document.getElementById("---ELID---_" + tabName).style.display = "block";  

          tabs = document.getElementsByClassName("tab_---ELID---");
          for (i = 0; i < x.length; i++)
            tabs[i].className = tabs[i].className.replace(" lvseltab", "");
          el.className += " lvseltab";
        }
        </script>
        """
        if self.buttons:
            html += "<div class='lvtabbar'>\n"
            for t in range(len(self.tabs)):
                #Add header items
                classes = 'lvbutton lvctrl tab_---ELID---'
                if t == 0: classes += ' lvseltab'
                html += '<a class="' + classes + '" onclick="openTab_---ELID---(this, this.innerHTML)">---LABEL---</a>'
                html = html.replace('---LABEL---', self.tabs[t])
            html += "</div>\n"
        for t in range(len(self.tabs)):
            #Add control wrappers
            style = ''
            if t != 0: style = 'display: none;'
            html += '<div id="---ELID---_---LABEL---" style="' + style + '" class="lvtab lvctrl ---ELID---">\n'
            for ctrl in self._content:
                if ctrl.tab == t:
                    html += ctrl.controls()
            html += '</div>\n'
            html = html.replace('---LABEL---', self.tabs[t])
        html = html.replace('---ELID---', self.elid)
        return html

class _Control(_HTML):
    """
    _Control object

    Parameters
    ----------
    target : Obj or Viewer
        Add a control to an object to control object properties
        Add a control to a viewer to control global proerties and run commands
    property : str
        Property to modify
    command : str
        Command to run
    value : any type
        Initial value of the controls
    label : str
        Descriptive label for the control
    readproperty : str
        Property to read control value from on update (but not modified)
    """

    def __init__(self, target, property=None, command=None, value=None, label=None, index=None, readproperty=None):
        super(_Control, self).__init__(label)
        self.label = label

        #Can either set a property directly or run a command
        self.property = readproperty
        self.target = target
        self.action = None
        if property:
            #Property set
            self.action = _PropertyAction(target, property, command, index)
            if label is None:
                self.label = property.capitalize()
            self.property = property
        elif command:
            #Command only
            self.action = _CommandAction(target, command, readproperty)
            if label is None:
                self.label = command.capitalize()
        else:
            #Assume derived class will fill out the action, this is just a placeholder
            self.action = _Action(target)

        if not self.label:
            self.label = ""

        #Get value from target or default if not provided
        if property is not None:
            if value is None:
                value = _getproperty(target, property)
            else:
                #2d/3D value?
                if index is not None:
                    V = target[property]
                    V[index] = value
                    target[property][index] = V #Set the provided value
                else:
                    target[property] = value #Set the provided value

            #Append reload command from prop dict if no command provided
            _lv = _getviewer(target)
            if property in _lv.properties:
                prop = _lv.properties[property]
                #TODO: support higher reload values
                cmd = ""
                if prop["redraw"] > 1 and not "reload" in str(command):
                    cmd = "reload"
                elif prop["redraw"] > 0 and not "redraw" in str(command):
                    cmd = "reload"
                #Append if command exists
                if command is None:
                    command = cmd
                else:
                   #Pass the value to command if provided
                   command += " ---VAL--- ; " + cmd

        self.value = value
        self._value = value #Store passed initial value

    def onchange(self):
        return "_wi['---VIEWERID---'].do_action('" + str(self.id) + "', this.value, this);"

    def show(self):
        #Show only this control
        html = '<div style="" class="lvctrl">\n'
        html += self.html()
        html += '</div>\n'

    def html(self):
        return self.controls()

    def controls(self, type='number', attribs={}, onchange=""):
        #Input control
        html =  '<input id="---ELID---_{0}" class="---ELID---" type="{0}"'.format(type)
        #Set custom attribute for property controls
        if not "step" in attribs or attribs["step"] is None:
            attribs["step"] = "any"
            #if self.value:
            #    attribs["step"] = abs(self.value) / 100.
            #else:
            #    attribs["step"] = "any"
        for key in attribs:
            html += key + '="' + str(attribs[key]) + '" '
        html += self.attribs()
        html += 'value="' + str(self.value) + '" '
        #Onchange event
        onchange += self.onchange()
        html += 'onchange="' + onchange + '" '
        html += '>\n'
        html = html.replace('---ELID---', self.elid)
        return html

    def attribs(self):
        html = ""
        if self.property:
            #Set custom attribute for property controls
            if _isobject(self.target):
                html += 'data-target="' + str(self.target["name"]) + '" '
            html += 'data-property="' + self.property + '" '
            if self.action.index is not None:
                html += 'data-index="' + str(self.action.index) + '" '
        return html

    def scripts(self):
        #Returns script dictionary
        return {self.id : self.action.script()}

class _MultiControl(_Control):
    """
    _Control object that holds multiple integrated Controls
    """

    def __init__(self, *args, **kwargs):
        super(_MultiControl, self).__init__(*args, **kwargs)
        self._content = []
        self._current = 0

    def add(self, ctrl):
        self._content.append(ctrl)
        return ctrl

    def controls(self):
        return self.html()

    def __iter__(self):
        return self

    def __next__(self):
        if self._current > len(self._content)-1:
            self._current = 0
            raise StopIteration
        else:
            self._current += 1
            return self._content[self._current-1]

    def html(self):
        html = self.labelhtml()
        for i in range(len(self._content)):
            html += self._content[i].controls()
        return html

    def scripts(self):
        #Returns script dictionary
        d = {}
        for i in range(len(self._content)):
            d[self._content[i].id] = self._content[i].action.script()
        return d

class Divider(_Control):
    """A divider element
    """

    def controls(self):
        return '<hr style="clear: both;">\n'

class Number(_Control):
    """A basic numerical input control
    """
    def __init__(self, target=None, property=None, command=None, value=None, label=None, index=None, step=None, readproperty=None):
        super(Number, self).__init__(target, property, command, value, label, index, readproperty)

        #Get step defaults from prop dict
        _lv = _getviewer(target)
        if step is None and property is not None and property in _lv.properties:
            prop = _lv.properties[property]
            #Check for integer type, set default step to 1
            T = prop["type"]
            ctrl = prop["control"]
            #Whole number step size?
            if "integer" in T:
                step = 1
            elif len(ctrl) > 1 and len(ctrl[1]) == 3:
                step = ctrl[1][2]
            elif "real" in T:
                r = 1.0
                #Use given range if any provided
                if len(ctrl) > 1 and len(ctrl[1]) > 1:
                    r = ctrl[1][1] - ctrl[1][0]
                step = r / 100.0

        self.step = step

    def controls(self):
        html = self.labelhtml()
        attribs = {"step" : self.step}
        html += super(Number, self).controls('number', attribs)
        return html + '<br>\n'

class Number2D(_MultiControl):
    """A set of two numeric controls for adjusting a 2D value
    """
    def __init__(self, target, property, label=None, step=None, *args, **kwargs):
        super(Number2D, self).__init__(target, property, label=label, *args, **kwargs)
        if self._value is None:
            self._value = _getproperty(target, property)
        self.ctrlX = self.add(Number(target=target, property=property, value=self._value[0], step=step, label="", index=0))
        self.ctrlY = self.add(Number(target=target, property=property, value=self._value[1], step=step, label="", index=1))

    def controls(self):
        html = self.labelhtml() + self.ctrlX.controls() + self.ctrlY.controls()
        return html.replace('<br>', '') + '<br>'

class Number3D(_MultiControl):
    """A set of three numeric controls for adjusting a 3D value
    """
    def __init__(self, target, property, label=None, step=None, *args, **kwargs):
        super(Number3D, self).__init__(target, property, label=label, *args, **kwargs)
        if self._value is None:
            self._value = _getproperty(target, property)
        self.ctrlX = self.add(Number(target=target, property=property, value=self._value[0], step=step, label="", index=0))
        self.ctrlY = self.add(Number(target=target, property=property, value=self._value[1], step=step, label="", index=1))
        self.ctrlZ = self.add(Number(target=target, property=property, value=self._value[2], step=step, label="", index=2))

    def controls(self):
        html = self.labelhtml() + self.ctrlX.controls() + self.ctrlY.controls() + self.ctrlZ.controls()
        return html.replace('<br>', '') + '<br>'

class Rotation(_MultiControl):
    """A set of six buttons for adjusting a 3D rotation
    """
    def __init__(self, target, *args, **kwargs):
        self.label="Rotation"
        super(Rotation, self).__init__(target, *args, **kwargs)

        self.ctrlX0 = self.add(Button(target=target, command="rotate x -1", label="-X"))
        self.ctrlX1 = self.add(Button(target=target, command="rotate x 1", label="+X"))
        self.ctrlY0 = self.add(Button(target=target, command="rotate y -1", label="-Y"))
        self.ctrlY1 = self.add(Button(target=target, command="rotate y 1", label="+Y"))
        self.ctrlZ0 = self.add(Button(target=target, command="rotate z -1", label="-Z"))
        self.ctrlZ1 = self.add(Button(target=target, command="rotate z 1", label="+Z"))

    def controls(self):
        html = self.labelhtml() + self.ctrlX0.controls() + self.ctrlX1.controls() + self.ctrlY0.controls() + self.ctrlY1.controls() + self.ctrlZ0.controls() + self.ctrlZ1.controls()
        return html.replace('<br>', '') + '<br>'


class Checkbox(_Control):
    """A checkbox control for a boolean value
    """
    def labelhtml(self):
        return '' #'<br>\n'

    def controls(self):
        attribs = {}
        if self.value: attribs = {"checked" : "checked"}
        html = "<label>\n"
        html += super(Checkbox, self).controls('checkbox', attribs)
        html += " " + self.label + "</label><br>\n"
        return html

    def onchange(self):
        return "; _wi['---VIEWERID---'].do_action('" + str(self.id) + "', this.checked ? 1 : 0, this);"

class Range(_Control):
    """A slider control for a range of values

    Parameters
    ----------
    range : list or tuple
        Min/max values for the range
    """
    def __init__(self, target=None, property=None, command=None, value=None, label=None, index=None, range=None, step=None, readproperty=None):
        super(Range, self).__init__(target, property, command, value, label, index, readproperty)

        #Get range & step defaults from prop dict
        _lv = _getviewer(target)
        defrange = [0., 1., 0.]
        T = None
        if property is not None and property in _lv.properties:
            prop = _lv.properties[property]
            #Check for integer type, set default step to 1
            T = prop["type"]
            if "integer" in T:
                defrange[2] = 1
            ctrl = prop["control"]
            if len(ctrl) > 1 and len(ctrl[1]) == 3:
                defrange = ctrl[1]

        if range is None:
            range = defrange[0:2]
        if step is None:
            step = defrange[2]

        self.range = range
        self.step = step
        if not step:
            #Whole number step size?
            r = range[1] - range[0]
            #Has no type from property, or type is not real: Assume a step size of 1 if range max-min > 5 and both are integers
            if (T is None or not "real" in T) and (r >= 5 and range[0] - int(range[0]) == 0 and range[1] - int(range[1]) == 0):
                self.step = 1
            #Range is less than 5 or range values indicate real numbers
            else:
                self.step = r / 100.0

    def controls(self):
        attribs = {"min" : self.range[0], "max" : self.range[1], "step" : self.step}
        html = self.labelhtml()
        html += super(Range, self).controls('number', attribs, onchange='this.nextElementSibling.value=this.value; ')
        html += super(Range, self).controls('range', attribs, onchange='this.previousElementSibling.value=this.value; ')
        return html + '<br>\n'

class Button(_Control):
    """A push button control to execute a defined command
    """
    def __init__(self, target, command, label=None):
        super(Button, self).__init__(target, None, command, None, label)

    def onchange(self):
        return "_wi['---VIEWERID---'].do_action('" + str(self.id) + "', '', this);"

    def labelhtml(self):
        return ''

    def controls(self):
        html = self.labelhtml()
        html =  '<input class="---ELID---" type="button" value="' + str(self.label) + '" '
        #Onclick event
        html += 'onclick="' + self.onchange() + '" '
        html += '><br>\n'
        html = html.replace('---ELID---', self.elid)
        return html

class Entry(_Control):
    """A generic input control for string values
    """
    def controls(self):
        html = self.labelhtml()
        html += '<input class="---ELID---" type="text" value="" '
        html += self.attribs()
        html += ' onkeypress="if (event.keyCode == 13) { _wi[\'---VIEWERID---\'].do_action(\'---ID---\', this.value.trim(), this); };"><br>\n'
        html = html.replace('---ELID---', self.elid)
        return html.replace('---ID---', self.id)

class Command(_Control):
    """A generic input control for executing command strings
    """
    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(command=" ", label="Command", *args, **kwargs)

    def controls(self):
        html = self.labelhtml()
        html += """
        <input class="---ELID---" type="text" value="" 
        onkeypress="if (event.keyCode == 13) { var cmd=this.value.trim(); 
        _wi['---VIEWERID---'].do_action('---ID---', cmd ? cmd : 'repeat', this); this.value=''; };"><br>\n
        """
        html = html.replace('---ELID---', self.elid)
        return html.replace('---ID---', self.id)

class File(_Control):
    """A file picker control

    Unfortunately there is no way to get the file path
    """
    def __init__(self, command="file", directory=False, multiple=False, accept="", *args, **kwargs):
        self.options = ""
        if directory:
            self.options += 'webkitdirectory directory '
        elif multiple:
            self.options += 'multiple '
        if accept:
            #Comma separated file types list, e.g. image/*,audio/*,video/*,.pdf
            self.options += 'accept="' + accept + '"'

        super(File, self).__init__(command=command, label="Load File", *args, **kwargs)

    def controls(self):
        html = self.labelhtml()
        html += """
        <div style="border:solid #aaa 1px; padding:10px; margin-bottom: 5px;">
        <input class="---ELID---" type="file" id="file_selector_---ELID---" ---OPTIONS--- name="files[]"/>
        <output id="list"></output>
        </div>
        <script>
          function fileSelected_---ELID---(event) {
            var output = [];
            for (var i = 0; i < event.target.files.length; i++) {
              var f = event.target.files.item(i);
              output.push('<li><strong>', escape(f.name), '</strong> (', f.type || 'n/a', ') - ',
                           f.size, ' bytes, last modified: ',
                           '</li>');
                _wi['---VIEWERID---'].do_action("---ID---", f.name);
            }
            document.getElementById('list').innerHTML = '<ul>' + output.join('') + '</ul>';
          }
          document.getElementById('file_selector_---ELID---').addEventListener('change', fileSelected_---ELID---, false);            
        </script>
        """
        html = html.replace('---ELID---', self.elid)
        html = html.replace('---OPTIONS---', self.options)
        return html.replace('---ID---', self.id)

class List(_Control):
    """A list of predefined input values to set properties or run commands

    Parameters
    ----------
    options: list
        List of the available value strings
    """
    def __init__(self, target, property=None, options=None, *args, **kwargs):
        #Get default options from prop dict
        if options is None:
            defoptions = []
            _lv = _getviewer(target)
            if property is not None and property in _lv.properties:
                prop = _lv.properties[property]
                ctrl = prop["control"]
                if len(ctrl) > 2 and len(ctrl[2]):
                    defoptions = ctrl[2]
            options = defoptions
        self.options = options
        super(List, self).__init__(target, property, *args, **kwargs)

    def controls(self):
        html = self.labelhtml()
        html += '<select class="---ELID---" id="---ELID---" value="" '
        html += self.attribs()
        html += 'onchange="' + self.onchange() + '">\n'
        for opt in self.options:
            #Each element of options list can be:
            # - dict {"label" : label, "value" : value, "selected" : True/False}
            # - list [value, label, selected]
            # - value only
            if isinstance(opt, dict):
                selected = "selected" if opt.selected else ""
                html += '<option value="' + str(opt["value"]) + '" ' + selected + '>' + opt["label"] + '</option>\n'
            elif isinstance(opt, list) or isinstance(opt, tuple):
                selected = "selected" if len(opt) > 2 and opt[2] else ""
                html += '<option value="' + str(opt[0]) + '" ' + selected + '>' + str(opt[1]) + '</option>\n'
            else:
                html += '<option>' + str(opt) + '</option>\n'
        html += '</select><br>\n'
        html = html.replace('---ELID---', self.elid)
        return html

class Colour(_Control):
    """A colour picker for setting colour properties
    """
    def __init__(self, *args, **kwargs):
        self.style = ""
        super(Colour, self).__init__(command="", *args, **kwargs)

    def controls(self):
        html = self.labelhtml()
        html += """
        <div class="colourbg checkerboard" style="---STYLE---">
          <div id="---ELID---" ---ATTRIBS--- class="colour ---ELID---" onclick="
            var col = new Colour(this.style.backgroundColor);
            var offset = [this.getBoundingClientRect().left, this.getBoundingClientRect().top];
            var el = this;
            var savefn = function(val) {
              var c = new Colour(0);
              c.setHSV(val);
              el.style.backgroundColor = c.html();
              _wi['---VIEWERID---'].do_action('---ID---', c.html(), el);
            }
            el.picker = new ColourPicker(savefn);
            el.picker.pick(col, offset[0], offset[1]);">
          </div>
        </div>
        <script>
        var el = document.getElementById("---ELID---");
        //Set the initial colour
        var col = new Colour('---VALUE---');
        el.style.backgroundColor = col.html();
        </script>
        """
        html = html.replace('---VALUE---', str(self.value))
        html = html.replace('---ELID---', self.elid)
        html = html.replace('---ATTRIBS---', self.attribs())
        html = html.replace('---STYLE---', self.style)
        return html.replace('---ID---', self.id)

class ColourIndicator(Colour):
    """A small indicator for showing/setting the colour of the object
    """
    def __init__(self, *args, **kwargs):
        super(ColourIndicator, self).__init__(*args, **kwargs)
        #Custom div style:
        self.style="border: #888 1px solid; display: inline-block; width: 20px; height: 20px;"
        self.label = ""
        #Set the colour from cached data so at least the indicated colour is correct
        if not self.target["colour"]:
            self.target["colour"] = self.target.colour.toString()

class Gradient(_Control):
    """A colourmap editor
    """
    def __init__(self, target, *args, **kwargs):
        super(Gradient, self).__init__(target, property="colourmap", command="", *args, **kwargs)
        #Get and save the map id of target object
        if _isviewer(target):
            raise(Exception("Gradient control requires an Object target, not Viewer"))
        self.maps = target.parent.state["colourmaps"]
        self.map = None
        for m in self.maps:
            if m["name"] == self.value:
                self.map = m
        self.selected = -1;

    def controls(self):
        html = self.labelhtml()
        html += """
        <canvas id="---ELID---_canvas" ---ATTRIBS--- width="512" height="24" class="palette checkerboard">
        </canvas>
        <script>
        var el = document.getElementById("---ELID---_canvas"); //Get the canvas
        //Store the maps
        el.colourmaps = ---COLOURMAPS---;
        el.currentmap = ---COLOURMAP---;
        el.selectedIndex = ---SELID---;
        if (!el.gradient) {
          //Create the gradient editor
          el.gradient = new GradientEditor(el, function(obj, id) {
            //Gradient updated
            //var colours = obj.palette.toJSON()
            el.currentmap = obj.palette.get(el.currentmap);
            _wi['---VIEWERID---'].do_action('---ID---', JSON.stringify(el.currentmap.colours));
            //_wi['---VIEWERID---'].do_action('---ID---', obj.palette.toJSON(), el);

            //Update stored maps list by name
            if (el.selectedIndex >= 0)
              el.colourmaps[el.selectedIndex].colours = el.currentmap.colours; //el.gradient.palette.get();
          }
          , true); //Enable premultiply
        }
        //Load the initial colourmap
        el.gradient.read(el.currentmap.colours);
        </script>
        """
        mapstr = json.dumps(self.maps)
        html = html.replace('---COLOURMAPS---', mapstr)
        if self.map:
            html = html.replace('---COLOURMAP---', json.dumps(self.map))
        else:
            html = html.replace('---COLOURMAP---', '"black white"')
        html = html.replace('---SELID---', str(self.selected))
        html = html.replace('---ELID---', self.elid)
        html = html.replace('---ATTRIBS---', self.attribs())
        return html.replace('---ID---', self.id)

class ColourMapList(List):
    """A colourmap list selector, populated by the default colour maps
    """
    def __init__(self, target, selection=None, *args, **kwargs):
        if _isviewer(target):
            raise(Exception("ColourMapList control requires an Object target, not Viewer"))
        #Load maps list
        if selection is None:
            selection = target.parent.defaultcolourmaps()
        options = [''] + selection
        #Also add the matplotlib colourmaps if available
        try:
            #Load maps list
            import matplotlib
            import matplotlib.pyplot as plt
            sel = matplotlib.pyplot.colormaps()
            options += matplotlib.pyplot.colormaps()
        except:
            pass

        #Preceding command with '.' calls via python API, allowing use of matplotlib maps
        super(ColourMapList, self).__init__(target, options=options, command=".colourmap", label="Load Colourmap", *args, **kwargs)

class ColourMaps(_MultiControl):
    """A colourmap list selector, populated by the available colour maps,
    combined with a colourmap editor for the selected colour map
    """
    def __init__(self, target, *args, **kwargs):
        #Load maps list
        if _isviewer(target):
            raise(Exception("ColourMaps control requires an Object target, not Viewer"))
        self.maps = target.parent.state["colourmaps"]
        options = [["", "None"]]
        sel = target["colourmap"]
        if sel is None: sel = ""
        for m in range(len(self.maps)):
            options.append([self.maps[m]["name"], self.maps[m]["name"]])
            #Mark selected
            if sel == self.maps[m]["name"]:
                options[-1].append(True)
                sel = m

        super(ColourMaps, self).__init__(target, *args, **kwargs)
        self.list = self.add(List(target, options=options, command="", property="colourmap", label="", *args, **kwargs))

        self.gradient = Gradient(target)
        self.gradient.selected = sel #gradient editor needs to know selection index
        self.gradient.label = self.label
        self.add(self.gradient) #Add control list

    def controls(self):
        html = self.list.controls() + self.gradient.controls()
        html = html.replace('---PALLID---', str(self.gradient.elid))
        return html + '<br>'

class TimeStepper(Range):
    """A time step selection range control with up/down buttons
    """
    def __init__(self, viewer, *args, **kwargs):
        #Acts as a command setter with some additional controls
        super(TimeStepper, self).__init__(target=viewer, label="Timestep", command="timestep", readproperty="timestep", *args, **kwargs)

        self.timesteps = viewer.timesteps()
        self.step = 1
        if len(self.timesteps):
            self.range = (self.timesteps[0], self.timesteps[-1])
            #Calculate step gap
            self.step = self.timesteps[1] - self.timesteps[0]
        else:
            self.range = (0,0)
        self.value = 0

    def controls(self):
        html = Range.controls(self)
        #Note: unicode symbol escape must use double slash to be
        # passed through to javascript or python will process them
        html += """
        <script>
        var timer_---ELID--- = -1;
        function startTimer_---ELID---() {
          if (timer_---ELID--- >= 0) {
            if (timer_---ELID--- > 0)
              window.cancelAnimationFrame(timer_---ELID---);
            timer_---ELID--- = window.requestAnimationFrame(nextStep_---ELID---);
          }
        }
        function nextStep_---ELID---() {
          el = document.getElementById('---ELID---_number');
          if (el) {
            //Call again on image load - pass callback
            var V = _wi['---VIEWERID---'];
            if (!V.box.canvas.mouse.isdown && !V.box.zoomTimer && (!V.box.gui || V.box.gui.closed))
              V.execute("next", startTimer_---ELID---);
            else
              setTimeout(nextStep_---ELID---, 100);
          }
        }
        function playPause_---ELID---(btn) {
          if (timer_---ELID--- >= 0) {
            btn.value="\\u25BA";
            window.cancelAnimationFrame(timer_---ELID---);
            timer_---ELID--- = -1;
          } else {
            timer_---ELID--- = 0;
            startTimer_---ELID---();
            btn.value="\\u25FE";
          }
        }
        </script>
        """
        html += '<input type="button" style="width: 50px;" onclick="var el = document.getElementById(\'---ELID---_range\'); el.stepDown(); el.onchange()" value="&larr;" />'
        html += '<input type="button" style="width: 50px;" onclick="var el = document.getElementById(\'---ELID---_range\'); el.stepUp(); el.onchange()" value="&rarr;" />'
        html += '<input type="button" style="width: 60px;" onclick="playPause_---ELID---(this);" value="&#9658;" />'
        html = html.replace('---ELID---', self.elid)
        return html

class DualRange(_MultiControl):
    """A set of two range slider controls for adjusting a minimum and maximum range

    Parameters
    ----------
    range : list or tuple
        Min/max values for the range
    """
    def __init__(self, target, properties, values=[None,None], label=None, range=(0.,1.), step=None):
        super(DualRange, self).__init__(target)
        self.label = label

        self.ctrlmin = self.add(Range(target=target, property=properties[0], step=step, value=values[0], range=range, label=""))
        self.ctrlmax = self.add(Range(target=target, property=properties[1], step=step, value=values[1], range=range, label=""))

class Range2D(_MultiControl):
    """A set of two range slider controls for adjusting a 2D value

    Parameters
    ----------
    range : list or tuple
        Min/max values for the ranges
    """
    def __init__(self, target, property, label=None, value=None, range=(0.,1.), step=None, *args, **kwargs):
        super(Range2D, self).__init__(target, property=property, command=None, value=value, label=label, *args, **kwargs)

        if self._value is None:
            self._value = _getproperty(target, property)

        self.ctrlX = self.add(Range(target=target, property=property, step=step, value=self._value[0], range=range, label="", index=0))
        self.ctrlY = self.add(Range(target=target, property=property, step=step, value=self._value[1], range=range, label="", index=1))

class Range3D(_MultiControl):
    """A set of three range slider controls for adjusting a 3D value

    Parameters
    ----------
    range : list or tuple
        Min/max values for the ranges
    """
    def __init__(self, target, property, label=None, value=None, range=(0.,1.), step=None, *args, **kwargs):
        super(Range3D, self).__init__(target, property=property, command=None, value=value, label=label, *args, **kwargs)

        if self._value is None:
            self._value = _getproperty(target, property)

        self.ctrlX = self.add(Range(target=target, property=property, step=step, value=self._value[0], range=range, label="", index=0))
        self.ctrlY = self.add(Range(target=target, property=property, step=step, value=self._value[1], range=range, label="", index=1))
        self.ctrlZ = self.add(Range(target=target, property=property, step=step, value=self._value[2], range=range, label="", index=2))

class Filter(DualRange):
    """A set of two range slider controls for adjusting a minimum and maximum filter range

    Parameters
    ----------
    range : list or tuple
        Min/max values for the filter range
    """
    def __init__(self, target, filteridx, label=None, range=None, step=None):
        self.label = label
        if len(target["filters"]) <= filteridx:
            print("Filter index out of range: ", filteridx, len(target["filters"]))
            return None
        self.filter = target["filters"][filteridx]

        #Default label - data set name
        if label is None:
            label = self.filter['by'].capitalize()

        #Get the default range limits from the matching data source
        fname = self.filter['by']
        if not target["data"]:
            #Attempt to update
            target.reload()
            target.parent.render()

        if not target["data"]:
            raise(ValueError("target data list empty, can't use filter"))
        if not fname in target['data']:
            raise(ValueError(fname + " not found in target data list, can't use filter"))
        self.data = target["data"][fname]
        if not range:
            #self.range = (self.filter["minimum"], self.filter["maximum"])
            if self.filter["map"]:
                range = (0.,1.)
            else:
                range = (self.data["minimum"]*0.99, self.data["maximum"]*1.01)

        #Setup DualRange using filter min/max
        super(Filter, self).__init__(_getviewer(target), properties=[None, None], values=[self.filter["minimum"], self.filter["maximum"]], label=label, range=range)

        #Replace actions on the controls
        self.ctrlmin.action = _FilterAction(target, filteridx, "minimum")
        self.ctrlmax.action = _FilterAction(target, filteridx, "maximum")

class ObjectList(_MultiControl):
    """A set of checkbox controls for controlling visibility of all visualisation objects
    """
    def __init__(self, viewer, *args, **kwargs):
        super(ObjectList, self).__init__(viewer, label="Objects", *args, **kwargs)
        for obj in viewer.objects.list:
            self.add(ColourIndicator(obj, "colour"))
            self.add(Checkbox(obj, "visible", label=obj["name"])) 

#TODO: broken
class ObjectSelect(_Container):
    """A list selector of all visualisation objects that can be used to
    choose the target of a set of controls

    Parameters
    ----------
    objects : list
        A list of objects to display, by default all available objects are added
    """
    def __init__(self, viewer, objects=None, *args, **kwargs):
        if not _isviewer(viewer):
            print("Can't add ObjectSelect control to an Object, must add to Viewer")
            return
        self.parent = viewer
        if objects is None:
            objects = viewer.objects.list
        
        #Load maps list
        options = [(0, "None")]
        for o in range(len(objects)):
            obj = objects[o]
            options += [(o+1, obj["name"])]

        #The list control
        self._list = List(target=viewer, label="", options=options, command="select", *args, **kwargs)

        #Init container
        super(ObjectSelect, self).__init__(viewer) #, label="Objects", options=options, command="select", *args, **kwargs)

        #Holds a control factory so controls can be added with this as a target
        self.control = _ControlFactory(self)

    #def onchange(self):
    #    #Update the control values on change
    #    #return super(ObjectSelect, self).onchange()
    #    return self._list.onchange()

    def __contains__(self, key):
        #print "CONTAINS",key
        obj = self._list.action.lastvalue
        #print "OBJECT == ",obj,(key in self.parent.objects.list[obj-1])
        return obj > 0 and key in self.parent.objects.list[obj-1]

    def __getitem__(self, key):
        #print "GETITEM",key
        obj = self._list.action.lastvalue
        if obj > 0:
            #Passthrough: Get from selected object
            return self.parent.objects.list[obj-1][key]
        return None

    def __setitem__(self, key, value):
        obj = self._list.action.lastvalue
        #print "SETITEM",key,value
        if obj > 0:
            #Passtrough: Set on selected object
            self.parent.objects.list[obj-1][key] = value

    #Undefined method call - pass call to target
    def __getattr__(self, key):
        #__getattr__ called if no attrib/method found
        def any_method(*args, **kwargs):
            #If member function exists on target, call it
            obj = self._list.action.lastvalue
            if obj > 0:
                method = getattr(self.parent.objects.list[obj-1], key, None)
                if method and callable(method):
                    return method(*args, **kwargs)
        return any_method

    def html(self):
        html = '<div style="border: #888 1px solid; display: inline-block; padding: 6px;" class="lvctrl">\n'
        html += self._list.controls()
        html += '<hr>\n'
        html += super(ObjectSelect, self).html()
        html += '</div>\n'
        return html

    def controls(self):
        return self.html()

class _ControlFactory(object):
    """
    Create and manage sets of controls for interaction with a Viewer or Object
    Controls can run commands or change properties
    """
    #Creates a control factory used to generate controls for a specified target
    def __init__(self, target):
        self._target = weakref.ref(target)
        self.clear()
        self.interactor = False
        #self.output = ""

        #Save types of all control/containers
        def all_subclasses(cls):
            return cls.__subclasses__() + [g for s in cls.__subclasses__() for g in all_subclasses(s)]

        #_Control contructor shortcut methods
        #(allows constructing controls directly from the factory object)
        #Use a closure to define a new method to call constructor and add to controls
        def addmethod(constr):
            def method(*args, **kwargs):
                #Return the new control and add it to the list
                newctrl = constr(self._target(), *args, **kwargs)
                self.add(newctrl)
                return newctrl
            return method

        self._control_types = all_subclasses(_Control)
        self._container_types = all_subclasses(_Container)
        for constr in self._control_types + self._container_types:
            key = constr.__name__
            method = addmethod(constr)
            #Set docstring (+ _Control docs)
            if constr in self._control_types:
                method.__doc__ = constr.__doc__ + _Control.__doc__
            else:
                method.__doc__ = constr.__doc__
            self.__setattr__(key, method)

    def __call__(self, property, *args, **kwargs):
        """
        Calling with a property name creates the default control for that property
        """
        _lv = _getviewer(self._target())
        if property is not None and property in _lv.properties:
            #Get control info from prop dict
            prop = _lv.properties[property]
            T = prop["type"]
            ctrl = prop["control"]
            if len(ctrl) > 2 and len(ctrl[2]) > 1:
                #Has selections
                return self.List(property, *args, **kwargs)
            if "integer" in T or "real" in T:
                hasrange = len(ctrl) > 1 and len(ctrl[1]) == 3
                #Multi-dimensional?
                if "[2]" in T:
                    return self.Range2D(property, *args, **kwargs) if hasrange else self.Number2D(property, *args, **kwargs)
                elif "[3]" in T:
                    return self.Range3D(property, *args, **kwargs) if hasrange else self.Number3D(property, *args, **kwargs)
                else:
                    return self.Range(property, *args, **kwargs) if hasrange else self.Number(property, *args, **kwargs)
            elif T == "string":
                return self.Entry(property, *args, **kwargs)
            elif T == "boolean":
                return self.Checkbox(property, *args, **kwargs)
            elif T == "colour":
                return self.Colour(property, *args, **kwargs)
            else:
                print("Unable to determine control type for property: " + property)
                print(prop)

    def add(self, ctrl):
        """
        Add a control
        """
        #Just add to the parent lists if not a viewer
        if not _isviewer(self._target()):
            self._target().parent.control.add(ctrl)
        else:
            if type(ctrl) in self._container_types:
                #New container, further controls will be added to it
                self._containers.append(ctrl)
            else:
                #Normal control: add to active container
                self._containers[-1].add(ctrl)

        return ctrl

    def show(self, menu=True, filename=None, fallback=None):
        """
        Displays all added controls including viewer if any

        Parameters
        ----------
        filename : str
            Filename to write output HTML
            If in a notebook this is not necessary as content is displayed inline
            If not provided and run outside IPython, defaults to "control.html"
        fallback : function
            A function which is called in place of the viewer display method when run outside IPython
        """
        #Show all controls in container
        target = self._target()

        #Creates an interactor to connect javascript/html controls to IPython and viewer
        #if no viewer Window() created, it will be a windowless interactor
        if _isviewer(target):
            #Append the current viewer ref
            windows.append(target)
            #Get unique ID
            winids.append(gen_id())
        else:
            target.parent.control.show()
            return

        viewerid = winids[-1]

        #Generate the HTML and associated action JS
        html = "<form novalidate>"
        chtml = ""
        actions = {}
        for con in self._containers:
            html += con.html()
            actions.update(con.scripts())
        html += "</form>"
        #print('\n'.join([str(a)+":"+actions[a] for a in actions]))

        #Auto-clear now content generated
        #Prevents doubling up if cell executed again
        self.clear()

        #Set viewer id
        html = html.replace('---VIEWERID---', str(viewerid))
        #self.output += html

        code = _webglboxcode(menu)

        #Display HTML inline or export
        if is_notebook() and filename is None:
            if not target.server:
                raise(Exception("LavaVu HTTP Server must be active for interactive controls, set port= parameter to > 0"))
            """
            HTTP server mode interaction, rendering in separate render thread:
             - This should work in all notebook contexts, colab, jupyterlab etc
             - Only problem remaining is port access, from docker or cloud instances etc, need to forward port
             - jupyter-server-proxy packaged (available on pip) supports forwarding port via the jupyter server
            """
            from IPython.display import display,HTML
            #Interaction requires some additional js/css/webgl
            #Insert stylesheet, shaders and combined javascript libraries
            display(HTML(code))

            #Try and prevent this getting cached
            timestamp = datetime.datetime.now().strftime('%Y%m%d%H%M%S')
            html += "<!-- CREATION TIMESTAMP {0} -->".format(timestamp)

            #Pass port and object id from server
            html += "<script>init('{0}');</script>".format(viewerid)
            actionjs = self.export_actions(actions, id(target), target.port)
            #Output the controls and start interactor
            display(HTML(actionjs + html))

        else:
            #Export standalone html
            #return as string or write output to .html file in current directory
            template = basehtml
            template = template.replace('---SCRIPTS---', code + self.export_actions(actions)) #Process actions
            template = template.replace('---INIT---', 'init(\'{0}\');'.format(viewerid))
            template = template.replace('---CONTENT---', html)
            full_html = template

            if filename == '':
                #Just return code
                return full_html
            else:
                #Write the file
                if filename is None:
                    filename = "control.html"
                hfile = open(filename, "w")
                hfile.write(full_html)
                hfile.close()

            if callable(fallback):
                fallback(target)

    def export_actions(self, actions, uid=0, port=0, proxy=False):
        #Process actions
        topjs = '<script type="text/javascript">\n'
        if port > 0:
            topjs += 'function init(viewerid) {{_wi[viewerid] = new WindowInteractor(viewerid, {uid}, {port});\n'.format(uid=uid, port=port)
        else:
            topjs += 'function init(viewerid) {{_wi[viewerid] = new WindowInteractor(viewerid, {uid});\n'.format(uid=uid)

        topjs += '_wi[viewerid].actions = {\n'

        actionjs = ''
        for act in actions:
            #Add to action functions to list, each takes the value of the control when called
            if len(actionjs):
                actionjs += ',\n'
            actionjs += '  "' + str(act) + '" : "' + actions[act] + '"'; 

        #Add init and finish
        return topjs + actionjs + '};\n}\n</script>\n'

    def redisplay(self):
        """Update the active viewer image if any
        Applies changes made in python to the viewer and forces a redisplay
        """
        if not is_notebook():
            return
        #Find matching viewer id, redisplay all that match
        for idx,obj in enumerate(windows):
            if obj == self._target():
                viewerid = winids[idx]
                from IPython.display import display,HTML,Javascript
                #display(Javascript('_wi["{0}"].redisplay();'.format(viewerid)))
                display(HTML('<script>_wi["{0}"].redisplay();</script>'.format(viewerid)))

    def update(self):
        """Update the control values from current viewer data
        Applies changes made in python to the UI controls
        """
        #NOTE: to do this now, all we need is to trigger a get_state call from interactor by sending any command
        #TODO: currently only works for controllers with a window, other controls will not be updated
        if not is_notebook():
            return
        #Find matching viewer id, update all that match
        for idx,obj in enumerate(windows):
            #if obj == self._target():
            viewerid = winids[idx]
            from IPython.display import display,HTML
            display(HTML('<script>_wi["{0}"].execute(" ");</script>'.format(viewerid)))
        
    def clear(self):
        #Initialise with a default control wrapper
        self._containers = [_Container(self)]




import os
import sys
from setuptools import setup
from setuptools.command.install import install
from setuptools.command.develop import develop
from setuptools.command.egg_info import egg_info
from setuptools import Extension
from setuptools import find_packages
import distutils
import distutils.sysconfig
import distutils.ccompiler
from distutils.errors import CompileError, LinkError
import subprocess
from multiprocessing import cpu_count
from ctypes.util import find_library
import platform
import glob
import shutil

#Current version
#(must be of the form X.Y.Z to trigger wheel builds)
version = "1.6.7"

"""
To release a new verison:

    1) Edit the version number above, then run below command (in a fresh checkout!).
    This will rebuild lib and docs in release mode with the new version
    (may need to reset notebooks to documentation versions)
    (docs repo is at: git@github.com:lavavu/Documentation.git)
    (may need: `cd docs/src; pip install -r requirements.txt`)

    >>> python setup.py check

    2) Check the docs, and make any further commits necessary to get them in order,
       and run above command again until docs are ready.

    Now ready for release, run this to commit and push changes to docs,
    then tag, commit and push new version.

    >>> python setup.py release

NOTE:
    To use particular libraries, set LV_LIB_DIRS and LV_INC_DIRS on command line before running, eg:
    LV_LIB_DIRS=${MESA_PATH/lib/x86_64-linux-gnu LV_INC_DIRS=${MESA_PATH}/include python setup.py install

"""

def write_version():
    """
    Writes version info to version.cpp
    """
    content = ""
    try:
        with open('src/version.cpp', 'r') as f:
            content = f.read()
    except:
        pass
    if not version in content:
        with open('src/version.cpp', 'w') as f:
            print("Writing new version: " + version)
            f.write('#include "version.h"\nconst std::string version = "%s";\n' % version)
    else:
        print("Version matches: " + version)

#Run with "new" to prepare for a new version after changing above
if sys.argv[-1] == 'check':
    write_version()
    os.system("LV_RELEASE=1 make -j$(nproc)")
    os.system("LV_RELEASE=1 make docs -B")
    import webbrowser
    webbrowser.open("docs/index.html")
    sys.exit()
#Run with "tag" arg to create a release tag
if sys.argv[-1] == 'release':
    #Commit and push docs
    wd = os.getcwd()
    os.chdir("docs")
    os.system("git add -u")
    os.system('git commit -m "Docs rebuilt for release"')
    os.system("git push")
    #Commit main repo
    os.chdir(wd)
    os.system("git add -u")
    os.system('git commit -m "Release version {0}"'.format(version))

    #Tag and push
    os.system("git tag -a {0} -m 'version {0}'".format(version))
    os.system("git push origin %s" % version)
    os.system("git push")
    sys.exit()

#Run with "publish" arg to upload the release
if sys.argv[-1] == 'publish':
    os.system("python setup.py sdist")
    os.system("twine upload dist/lavavu-%s.tar.gz" % version)
    sys.exit()

# Just output the wheel name for current arch + environment
if sys.argv[-1] == 'wheelname':
    # https://stackoverflow.com/a/60644659/866759
    from setuptools.dist import Distribution
    lavavu = Extension('lavavu', [])
    dist = Distribution(attrs={'name': 'lavavu', 'version': version, 'ext_modules': [lavavu]})
    bdist_wheel_cmd = dist.get_command_obj('bdist_wheel')
    bdist_wheel_cmd.ensure_finalized()
    distname = bdist_wheel_cmd.wheel_dist_name
    tag = '-'.join(bdist_wheel_cmd.get_tag())
    wheel_name = f'{distname}-{tag}.whl'
    print(wheel_name)
    sys.exit()

#Get extra lib and include dirs
inc_dirs = []
lib_dirs = []
if 'LV_LIB_DIRS' in os.environ:
    lib_dirs += os.environ['LV_LIB_DIRS'].split(':')
if 'LV_INC_DIRS' in os.environ:
    inc_dirs += os.environ['LV_INC_DIRS'].split(':')

def build_sqlite3(sqlite3_path):
    """Builds sqlite3 from included submodule
    """
    compiler = distutils.ccompiler.new_compiler()
    assert isinstance(compiler, distutils.ccompiler.CCompiler)
    distutils.sysconfig.customize_compiler(compiler)

    #Add any include dirs
    compiler.add_include_dir(sqlite3_path)

    try:
        obj = compiler.compile([os.path.join(sqlite3_path, 'sqlite3.c')], 'build')
        return obj
    except CompileError:
        print('sqlite3 compile error')
        raise

#From https://stackoverflow.com/a/28949827/866759
def check_libraries(libraries, headers, extra_lib_dirs=[], extra_inc_dirs=[]):
    """check if the C module can be built by trying to compile a small
    program against the passed library with passed headers"""

    import tempfile
    # write a temporary .c file to compile
    c_code = "int main(int argc, char* argv[]) { return 0; }"
    #Add headers
    for header in headers:
        c_code = "#include <" + header + ">\n" + c_code

    tmp_dir = tempfile.mkdtemp(prefix = 'tmp_comp__')
    bin_file_name = os.path.join(tmp_dir, 'test_comp')
    file_name = bin_file_name + '.c'
    with open(file_name, 'w') as fp:
        fp.write(c_code)

    # and try to compile it
    compiler = distutils.ccompiler.new_compiler()
    assert isinstance(compiler, distutils.ccompiler.CCompiler)
    distutils.sysconfig.customize_compiler(compiler)

    #Add any extra library or include dirs
    for l in lib_dirs + extra_lib_dirs:
        compiler.add_library_dir(l)
    for p in inc_dirs + extra_inc_dirs:
        compiler.add_include_dir(p)


    try:
        compiler.link_executable(
            compiler.compile([file_name]),
            bin_file_name,
            libraries=libraries,
        )
    except (CompileError) as e:
        print('Libraries ' + str(libraries) + ' test compile error', e)
        ret_val = False
    except (LinkError) as e:
        print('Libraries ' + str(libraries) + ' test link error', e)
        ret_val = False
    else:
        print('Libraries ' + str(libraries) + ' found and passed compile test')
        ret_val = True
    shutil.rmtree(tmp_dir)
    return ret_val

if __name__ == "__main__":
    print("Running setup with args:", sys.argv)
    #Update version.cpp
    write_version()

    sqlite3_path = os.path.join('src', 'sqlite3')
    if not os.path.exists(os.path.join(sqlite3_path, 'sqlite3.c')):
        #Attempt to get sqlite3 source submodule if not checked out
        os.system("git submodule update --init")

    _debug = False
    #_debug = True
    srcs = [os.path.join('src', 'LavaVuPython_wrap.cxx')]
    srcs += glob.glob(os.path.join('src', '*.cpp'))
    srcs += glob.glob(os.path.join('src', 'Main', '*.cpp'))
    srcs += glob.glob(os.path.join('src', 'jpeg', '*.cpp'))
    defines = [('USE_FONTS', '1')]
    cflags = []
    libs = []
    ldflags = []
    rt_lib_dirs = []
    extra_objects = []
    try:
        import numpy
    except:
        subprocess.check_call([sys.executable, '-m', 'pip', 'install', 'numpy>=1.11.0'])
        try:
            import numpy
        except:
            subprocess.check_call([sys.executable, '-m', 'pip3', 'install', 'numpy>=1.11.0'])
            import numpy
    inc_dirs += [numpy.get_include()]

    P = platform.system()
    if _debug:
        print("DEBUG BUILD ENABLED")
        defines += [('DEBUG', '1')]
        if P == 'Windows':
            cflags += ['/Zi']
            ldflags += ['/DEBUG']
        else:
            cflags += ['-g', '-O0']

    #Parallel build
    num_jobs = 1
    if not 'arm' in platform.machine():
        try:
            num_jobs = os.cpu_count()
        except AttributeError:
            import multiprocessing
            num_jobs = multiprocessing.cpu_count()

    if num_jobs > 1:
        try:
            from numpy.distutils.ccompiler import CCompiler_compile
            import distutils.ccompiler
            distutils.ccompiler.CCompiler.compile = CCompiler_compile
            os.environ['NPY_NUM_BUILD_JOBS'] = str(num_jobs)
            print('Using parallel build, jobs: ', num_jobs)
        except ImportError:
            print("Numpy not found, parallel compile not available")

    #OS Specific
    if P == 'Windows':
        #Windows - includes all dependencies

        #VC++ can handle C or C++ without issues clang has,
        #so just throw sqlite source in with the rest
        inc_dirs += [sqlite3_path]
        srcs += [os.path.join(sqlite3_path, 'sqlite3.c')]

        #32 or 64 bit python interpreter?
        if sys.maxsize > 2**32:
            arc = '64'
        else:
            arc = '32'

        #Download, extract and install ffmpeg files
        #(NOTE: sdist on windows is currently useless as does not include dlls,
        #a fix would need to download other dll's here as we do for ffmpeg)
        ffmpeg_dlls = []
        ffmpeg_libs = []
        try:
            sys.path.append('lavavu')
            import vutils
            import zipfile
            src = "https://github.com/OKaluza/ffmpeg.zeranoe.com/raw/main/"
            fn1 = 'ffmpeg-4.3.1-win' + arc + '-shared'
            fn2 = 'ffmpeg-4.3.1-win' + arc + '-dev'
            LIBS = 'lib' + arc
            outfn = vutils.download(src + fn1 + '.zip')
            with zipfile.ZipFile(outfn, 'r') as zip_ref:
                zip_ref.extractall('.')
            outfn = vutils.download(src + fn2 + '.zip')
            with zipfile.ZipFile(outfn, 'r') as zip_ref:
                zip_ref.extractall('.')
            #Now copy into windows build dirs
            libs = ['avformat', 'avcodec', 'avutil', 'swscale']
            for lib in libs:
                #Headers - move entire directories
                dst = os.path.join('src', 'windows', 'inc', 'lib' + lib)
                if not os.path.exists(dst):
                    shutil.move(os.path.join(fn2, 'include', 'lib' + lib), dst)
                #Lib file
                dst = os.path.join('src', 'windows', LIBS)
                shutil.copy(os.path.join(fn2, 'lib', lib + '.lib'), dst)
                ffmpeg_libs += [lib]
            #Just grab all the dlls
            ffmpeg_dlls = glob.glob(os.path.join(fn1, 'bin', '*.dll'))

            #If we got this far, enable video
            defines += [('HAVE_LIBAVCODEC', '1'), ('HAVE_SWSCALE', '1')]

        except (Exception) as e:
            print("ffmpeg download/extract failed", str(e))
            ffmpeg_dlls = []
            ffmpeg_libs = []
            pass

        srcs += [os.path.join('src', 'png', 'lodepng.cpp')]
        srcs += [os.path.join('src', 'miniz', 'miniz.c')]
        defines += [('HAVE_GLFW', '1')]
        #defines += [('HAVE_LIBPNG', 1)]
        #inc_dirs += [os.path.join(os.getcwd(), 'src', 'windows', 'inc')]
        #lib_dirs += [os.path.join(os.getcwd(), 'src', 'windows', LIBS)]
        inc_dirs += [os.path.join('src', 'windows', 'inc')]
        lib_dirs += [os.path.join('src', 'windows', LIBS)]
        #ldflags += ['/LIBPATH:' + os.path.join(os.getcwd(), 'src', 'windows', LIBS)]
        ldflags += ['/LIBPATH:' + os.path.join('src', 'windows', LIBS)]
        libs += ['opengl32', 'pthreadVC2', 'glfw3dll'] + ffmpeg_libs
        dlls = [os.path.join('src', 'windows', LIBS, 'pthreadVC2.dll'),
                os.path.join('src', 'windows', LIBS, 'glfw3.dll')] + ffmpeg_dlls
        #Copy dlls into ./lavavu so can be found by package_data
        for d in dlls:
            shutil.copy(d, 'lavavu')
    else:
        #POSIX only - find external dependencies
        cflags += ['-std=c++0x']
        #Use external sqlite3 if available, otherwise build
        if not (find_library('sqlite3') and check_libraries(['sqlite3'], ['sqlite3.h'], extra_inc_dirs=[sqlite3_path])):
            #Don't compile when just running 'setup.py egg_info'
            if len(sys.argv) < 2 or sys.argv[1] == 'egg_info':
                sqlite3 = 'build/src/sqlite3/sqlite3.o'
            else:
                sqlite3 = build_sqlite3(sqlite3_path)
            inc_dirs += [sqlite3_path]
            extra_objects = sqlite3
        else:
            libs += ['sqlite3']

        # Optional external libraries - check if installed
        if find_library('png') and check_libraries(['png', 'z'], ['png.h', 'zlib.h']):
            defines += [('HAVE_LIBPNG', 1), ('USE_ZLIB', '1')]
            libs += ['png', 'z']
        else:
            srcs += ['src/png/lodepng.cpp']
            if find_library('z') and check_libraries(['z'], ['zlib.h']):
                defines += [('USE_ZLIB', '1')]
                libs += ['z']
            else:
                srcs += ['src/miniz/miniz.c']

        if find_library('tiff') and check_libraries(['tiff'], ['tiffio.h']):
            defines += [('HAVE_LIBTIFF', 1)]
            libs += ['tiff']

        inc_dirs += ['/usr/include/ffmpeg'] #Required for centos/redhat ffmpeg-devel
        if (find_library('avcodec') and find_library('avformat') and find_library('avutil')
            and check_libraries(['avcodec', 'avformat', 'avutil'],
                ['libavformat/avformat.h', 'libavcodec/avcodec.h', 'libavutil/mathematics.h',
                 'libavutil/imgutils.h'])):
            defines += [('HAVE_LIBAVCODEC', 1)]
            libs += ['avcodec', 'avformat', 'avutil']
            if find_library('swscale') and check_libraries(['swscale'], ['libswscale/swscale.h']):
                defines += [('HAVE_SWSCALE', 1)]
                libs += ['swscale']

        if P == 'Linux':
            #Linux GLFW, X11, EGL or OSMesa
            #To force EGL or OSMesa, set LV_EGL=1 or LV_OSMESA=1
            if ("LV_EGL" in os.environ and find_library('OpenGL') and find_library('EGL')
            and check_libraries(['OpenGL', 'EGL'], ['GL/gl.h', 'EGL/egl.h'])):
                #EGL for offscreen OpenGL without X11/GLX - works only with NVidia currently
                defines += [('HAVE_EGL', '1')]
                libs += ['OpenGL', 'EGL']
            elif ("LV_OSMESA" in os.environ and find_library('OSMesa')
            and check_libraries(['OSMesa'], ['GL/osmesa.h'])):
                #OSMesa for software rendered offscreen OpenGL
                defines += [('HAVE_OSMESA', '1')]
                libs += ['OSMesa']
            elif ("LV_GLFW" in os.environ and find_library('glfw')
            and check_libraries(['glfw'], ['GLFW/glfw3.h'])):
                #Use GLFW
                defines += [('HAVE_GLFW', '1')]
                libs += ['GL', 'glfw']
            else:
                #Fallback - X11
                defines += [('HAVE_X11', '1')]
                libs += ['GL', 'X11']

            #Runtime libraries: set rpath: $ORIGIN includes current dir in search
            rt_lib_dirs += ['$ORIGIN'] + lib_dirs

        elif P == 'Darwin':
            #Mac OS X with Cocoa + CGL
            #srcs += ['src/Main/CocoaViewer.mm']
            #This hack is because setuptools can't handle .mm extension
            shutil.copyfile('src/Main/CocoaViewer.mm', 'src/Main/CocoaViewer.m')
            srcs += ['src/Main/CocoaViewer.m']
            cflags += ['-ObjC++'] #Now have to tell compiler it's objective c++ as .m file indicates objective c
            defines += [('HAVE_CGL', '1'), ('GL_SILENCE_DEPRECATION','1')]
            cflags += ['-undefined suppress', '-flat_namespace'] #Swig, necessary?
            cflags += ['-Wno-unknown-warning-option', '-Wno-c++14-extensions', '-Wno-shift-negative-value']
            cflags += ['-FCocoa', '-FOpenGL', '-stdlib=libc++']
            libs += ['c++', 'objc']
            #Can't use extra_link_args because they are appended by setuptools but frameworks must come first
            os.environ['LDFLAGS'] = '-framework Cocoa -framework Quartz -framework OpenGL'
            #Runtime library dirs doesn't work on mac, so set rpath manually
            for l in lib_dirs:
                ldflags.append('-Wl,-rpath,'+l)

        #Other posix libs
        libs += ['dl', 'pthread', 'm']

    lv = Extension('lavavu._LavaVuPython',
                    define_macros = defines,
                    include_dirs = inc_dirs,
                    libraries = libs,
                    library_dirs = lib_dirs,
                    runtime_library_dirs = rt_lib_dirs,
                    extra_compile_args = cflags,
                    extra_link_args = ldflags,
                    extra_objects=extra_objects,
                    sources = srcs)

    #For some reason, jupyter_server_proxy requirement breaks conda build
    requirements = ['numpy>=1.11', 'aiohttp', 'jupyter_server_proxy']
    if 'CONDA_BUILD' in os.environ:
        del requirements[-1]

    #Binary package data for wheels
    #Package_data works for wheels(binary) only - so add everything we need for the wheels here
    #(MANIFEST.in contents will be included with sdist only unless include_package_data is enabled)
    package_data = {'lavavu': ['font.bin', 'dict.json', 'shaders/*', 'html/*']}
    if P == 'Windows':
        package_data['lavavu'] += ['*.dll']

    #If built with libOSMesa, rename package so can be distributed as an alternative
    dist_name = "lavavu"
    if "OSMesa" in libs:
        dist_name = "lavavu-osmesa"

    setup(name = dist_name,
          author            = "Owen Kaluza",
          author_email      = "owen.kaluza@monash.edu",
          url               = "https://github.com/lavavu/LavaVu",
          version           = version,
          license           = "LGPL-3",
          description       = "Python interface to LavaVu OpenGL 3D scientific visualisation utilities",
          long_description  = 'See https://github.com/lavavu/LavaVu for more info',
          packages          = ['lavavu'],
          package_dir       = {'lavavu': 'lavavu'},
          package_data      = package_data,
          install_requires  = requirements,
          platforms         = ['any'],
          entry_points      = {
              'gui_scripts': [
                  'LV = lavavu.__main__:main',
                  'LavaVu = lavavu.__main__:main'
              ]
          },
          #Use below to include files from MANIFEST.in or specify package_data, not both
          #include_package_data = True,
          classifiers = [
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
            'Operating System :: POSIX :: Linux',
            'Operating System :: MacOS',
            'Operating System :: Microsoft :: Windows',
            'Environment :: X11 Applications',
            'Environment :: MacOS X :: Cocoa',
            'Environment :: Win32 (MS Windows)',
            'Topic :: Multimedia :: Graphics :: 3D Rendering',
            'Topic :: Scientific/Engineering :: Visualization',
            'Development Status :: 5 - Production/Stable',
            'Programming Language :: C++',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Framework :: Jupyter',
            'Framework :: IPython',
          ],
          ext_modules = [lv]
          )




from __future__ import print_function
"""
Warning! EXPERIMENTAL:
 these features and functions are under development, will have bugs,
 and may be heavily modified in the future

Tools for converting between 3D data types

- Points to Volume
- Triangles to OBJ file
"""
import numpy
import os
import sys

def min_max_range(verts):
    """
    Get bounding box from a list of vertices
    returns (min, max, range)
    """
    vmin = numpy.min(verts, axis=0)
    vmax = numpy.max(verts, axis=0)
    vrange = vmax - vmin
    print("Bounding box ", (vmin, vmax), "Range ", vrange)
    return (vmin, vmax, vrange)

def default_sample_grid(vrange, res=8):
    """Calculate sample grid fine enough to capture point details
    resolution of smallest dimension will be minres elements
    """
    #If provided a full resolution, use that, otherwise will be interpreted as min res
    if len(list(res)) == 3:
        return list(res)
    #Use bounding box range min
    minr = numpy.min(vrange)
    #res parameter is minimum resolution
    factor = float(res) / float(minr)
    RES = [int(factor*(vrange[0])), int(factor*(vrange[1])), int(factor*(vrange[2]))]   
    print("Sample grid RES:",RES)
    return RES

def points_to_volume(verts, res=8, kdtree=False, normed=False, clamp=None, boundingbox=None):
    """
    Convert object vertices to a volume by interpolating points to a grid

    - Vertex data only is used, so treated as a point cloud
    - Can also be used to sample an irregular grid to a regular so volume render matches the grid dimensions
    - Result is a density field that can be volume rendered

    Default is to use numpy.histogramdd method, pass kdtree=True to use scipy.spatial.KDTree

    TODO: support colour data too, converted density field becomes alpha channel

    Returns
    -------
    values: numpy array of float32
        The converted density field
    boundingbox : 2,3
        the minimum 3d vertex of the bounding box
        the maximum 3d vertex of the bounding box

    """
    if kdtree:
        return points_to_volume_tree(verts, res)
    else:
        return points_to_volume_histogram(verts, res, normed, clamp, boundingbox)

def points_to_volume_histogram(verts, res=8, normed=False, clamp=None, boundingbox=None):
    """
    Using numpy.histogramdd to create 3d histogram volume grid
    (Easily the fastest, but less control over output)
    """
    #Reshape to 3d vertices
    verts = verts.reshape((-1,3))

    #Get bounding box of swarm
    vmin, vmax, vrange = min_max_range(verts)

    #Minimum resolution to get ok sampling
    if boundingbox is not None:
        vmin = boundingbox[0]
        vmax = boundingbox[1]
        vrange = numpy.array(boundingbox[1]) - numpy.array(boundingbox[0])
    RES = default_sample_grid(vrange, res)

    #H, edges = numpy.histogramdd(verts, bins=RES)
    if boundingbox is None:
        H, edges = numpy.histogramdd(verts, bins=RES, normed=normed) #density=True for newer numpy
    else:
        rg = ((vmin[0], vmax[0]), (vmin[1], vmax[1]), (vmin[2], vmax[2])) #provide bounding box as range
        H, edges = numpy.histogramdd(verts, bins=RES, range=rg, normed=normed) #density=True for newer numpy

    #Reverse ordering X,Y,Z to Z,Y,X for volume data
    values = H.transpose()

    #Clamp [0,0.1] - optional (for a more binary output, points vs no points)
    if clamp is not None:
        values = numpy.clip(values, a_min=clamp[0], a_max=clamp[1])
    
    #Normalise [0,1] if using probability density
    if normed:
        values = values / numpy.max(values)
    
    return (values, vmin, vmax)

def points_to_volume_tree(verts, res=8):
    """
    Using scipy.spatial.KDTree to find nearest points on grid

    Much slower, but more control

    TODO: control parameters
    """
    #Reshape to 3d vertices
    verts = verts.reshape((-1,3))

    #Get bounding box of swarm
    lmin, lmax, lrange = min_max_range(verts)
    print("Volume bounds: ",lmin.tolist(), lmax.tolist(), lrange.tolist())
    
    #Push out the edges a bit, will create a smoother boundary when we filter
    #lmin -= 0.1*lrange
    #lmax += 0.1*lrange
    values = numpy.full(shape=(verts.shape[0]), dtype=numpy.float32, fill_value=1.0)

    #Minimum resolution to get ok sampling
    RES = default_sample_grid(lrange, res)

    #Push out the edges a bit, will create a smoother boundary when we filter
    cell = lrange / RES #Cell size
    print("CELL:",cell)
    lmin -= 2.0*cell
    lmax += 2.0*cell

    x = numpy.linspace(lmin[0], lmax[0] , RES[0])
    y = numpy.linspace(lmin[1], lmax[1] , RES[1])
    z = numpy.linspace(lmin[2], lmax[2] , RES[2])
    print(x.shape, y.shape, z.shape)
    Z, Y, X = numpy.meshgrid(z, y, x, indexing='ij') #ij=matrix indexing, xy=cartesian
    XYZ = numpy.vstack((X,Y,Z)).reshape([3, -1]).T

    print(lmin,lmax, XYZ.shape)

    #KDtree for finding nearest neighbour points
    from scipy import spatial
    #sys.setrecursionlimit(1000)
    #tree = spatial.KDTree(XYZ)
    print("Building KDTree")
    tree = spatial.cKDTree(XYZ)

    #Outside distance to apply to grid points
    MAXDIST = max(lrange) / max(RES) #Max cell size diagonal
    print("Tree query, maxdist:",MAXDIST,max(lrange))

    #Query all points, result is tuple: distances, indices
    distances, indices = tree.query(verts, k=1) #Just get a single nearest neighbour for now

    print("Calculate distances")
    #Convert distances to [0,1] where 1=on grid and <= 0 is outside max range
    distances = (MAXDIST - distances) / MAXDIST
    print("Zero out of range distances")
    distances *= (distances>0) #Zero negative elements by multiplication in-place

    #Add the distances to the values at their nearest grid point indices
    print("Compose distance field")
    values = numpy.zeros(shape=(XYZ.shape[0]), dtype=numpy.float32)
    values[indices] += distances
    #Clip value max
    print("Clip distance field")
    values = values.clip(max=1.0)
    
    #Reshape to actual grid dims Z,Y,X... (not required but allows slicing)
    XYZ = XYZ.reshape(RES[::-1] + [3])
    values = values.reshape(RES[::-1])
    print(XYZ.shape, values.shape)
    
    return (values, lmin, lmax)


def points_to_volume_3D(vol, objects, res=8, kdtree=False, blur=0, pad=None, normed=False, clamp=None):
    """
    Interpolate points to grid and load into passed volume object

    Given a list of objects and a volume object, convert a point cloud
    from another object (or objects - list is supported) into a volume using
    points_to_volume()
    """
    lv = vol.parent #Get the viewer from passed object
    lv.hide(objects) #Hide the converted objects

    #Get vertices from lavavu objects
    pverts, bb_all = lv.get_all_vertices(objects)

    #blur = False
    #Use bounding box of full model?
    #vmin, vmax, vrange = min_max_range([lv["min"], lv["max"]])
    #bb_all == (vmin, vmax)
    vdata, vmin, vmax = points_to_volume(pverts, res, kdtree, normed, clamp, bb_all)

    if blur > 0:
        if pad == None:
            pad = int(blur*2)
        if pad > 0:
            print("Pad edges before blur", pad)
            vdata = numpy.pad(vdata, pad, mode='constant')
        print("Filter/blur distance field, sigma=%d" % blur)
        from scipy.ndimage.filters import gaussian_filter
        values = gaussian_filter(vdata, sigma=blur) #, mode='nearest')
    else:
        values = vdata #Need extra space at edges to use blur, so skip, or use pad() 
    
    print(numpy.min(values), numpy.max(values))
    print(numpy.min(values), numpy.max(values))
    
    vol.values(values)
    vol.vertices((vmin, vmax))

def points_to_volume_4D(vol, objects, res=8, kdtree=False, blur=False, normed=False, clamp=None):
    """
    Interpolate points to grid at each timestep

    Given a list of objects and a volume object, convert a time-varying point cloud
    from another object (or objects - list is supported) into a volume using
    points_to_volume_3D()
    """
    lv = vol.parent #Get the viewer from passed object

    for step in lv.steps:
        print("TIMESTEP:",step)
        lv.timestep(step)
        
        points_to_volume_3D(vol, objects, res, kdtree, blur, normed, clamp)

def colour2rgb(c):
    return [c & 255, (c >> 8) & 255, (c >> 16) & 255]

def colour2hex(rgb):
    def padhex2(i):
        s = hex(int(i))
        return s[2:].zfill(2)
    return "#" + padhex2(rgb[0]) + padhex2(rgb[1]) + padhex2(rgb[2])

def _get_objects(source):
    """
    Returns a list of objects

    If source is lavavu.Viewer() list contains all objects
    If source is lavavu.Object() list contains that single object
    """
    if source.__class__.__name__ == 'Viewer':
        return source.objects.list
    elif not isinstance(source, list):
        return [source]
    else:
        return source

def export_OBJ(filepath, source, verbose=False):
    """
    Export given object(s) to an OBJ file
    Supports only triangle mesh object data

    If source is lavavu.Viewer() exports all objects
    If source is lavavu.Object() exports single object
    """
    mtlfilename = os.path.splitext(filepath)[0] + '.mtl'
    objects = _get_objects(source)
    with open(filepath, 'w') as f, open(mtlfilename, 'w') as m:
        f.write("# OBJ file\n")
        offset = 1
        for obj in objects:
            if obj["visible"]:
                f.write("g %s\n" % obj.name)
                offset = _write_OBJ(f, m, filepath, obj, offset, verbose)

def _write_MTL(m, name, texture=None, diffuse=[1.0, 1.0, 1.0], ambient=None, specular=None, opacity=1.0, illum=None):
    #http://paulbourke.net/dataformats/mtl/
    #print("Writing MTL ", texture, diffuse, opacity, name)
    mtl = "newmtl %s\n" % name
    mtl += "Kd %.06f %.06ff %.06f\n" % (diffuse[0], diffuse[1], diffuse[2])
    if ambient:
        mtl += "Ka %.06f %.06ff %.06f\n" % (ambient[0], ambient[1], ambient[2])
    if specular:
        if illum is None:
            illum = 2 #Highlight on
        mtl += "Ks %.06f %.06ff %.06f\n" % (specular[0], specular[1], specular[2])
        if len(specular) > 3:
            mtl += "Ns %.06f\n" % specular[3]
    mtl += "d %f\n" % (1.0 - opacity) #Dissolve: inverse of opacity
    if illum is None:
        illum = 1 #Default=1 = colour on, ambient on
    mtl += "illum %d\n" % illum

    if texture:
        mtl += "map_Kd %s\n" % texture

    mtl += '\n'
    m.write(mtl)
    mtl_line = "usemtl %s\n" % name
    return mtl_line

def _write_OBJ(f, m, filepath, obj, offset=1, verbose=False):
    mtl_line = ""
    cmaptexcoords = []
    colourdict = None
    import re
    name = re.sub(r"\s+", "", obj["name"])
    colourcount = sum([len(c) for c in obj.data.colours])
    if m and obj["texture"]:
        fn = obj["texture"]
        if fn == 'colourmap':
            fn = 'texture.png'
        #Write palette.png
        obj.parent.palette('image', 'texture')
        if verbose:
            print("Writing texture mtl ", fn)
        mtl_line = _write_MTL(m, name, texture=fn, opacity=obj["opacity"])

    elif m and colourcount > 0:
        #"""
        #18bit RGB = 262144 colours
        #Top-left = (0,0,0)
        #Bottom-right = (255,255,255)
        #(8x8 tiles of 64x64)
        #G - slow, Z index = G
        #R - midd, Y axis = R
        #B - fast, X axis = B
        #X = B//4
        #Y = R//4
        #Z = G//4
        #zx = Z % 8
        #zy = Z // 8
        #x = zx * 64 + X
        #y = zy * 64 + Y
        #u = x / 512
        #v = y / 512
        #https://upload.wikimedia.org/wikipedia/commons/3/34/RGB_18bits_palette.png
        palimg = u"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"
        import base64
        decoded = base64.b64decode(palimg).decode('utf-8')
        with open("palette_rgb.png", "wb") as f:
            f.write(decoded.encode('ascii'))
        mtl_line = _write_MTL(m, "palette_rgb", texture="palette_rgb.png", opacity=obj["opacity"])
        colourdict = {}
        """
        #SLOW!
        print("Creating palette")
        colourdict = {}
        for data in obj:
            for c in data.colours:
                rgb = colour2rgb(c)
                cs = colour2hex(rgb)
                colourdict[cs] = rgb
        print("Writing mtl lib (colour list)")
        for c in colourdict:
            rgb = colourdict[c]
            mtl_line = _write_MTL(m, c[1:], diffuse=[rgb[0]/255.0, rgb[1]/255.0, rgb[2]/255.0], opacity=obj["opacity"])
        """

    elif m and "colour" in obj:
        #print("Writing mtl lib (default colour)")
        c = obj.parent.parse_colour(obj["colour"])
        mtl_line = _write_MTL(m, 'default-' + name, diffuse=c, opacity=obj["opacity"])

    for o,data in enumerate(obj):
        if verbose: print("[%s] element %d of %d" % (obj.name, o+1, len(obj.data.vertices)))
        verts = data.vertices.reshape((-1,3))
        if len(verts) == 0:
            if verbose: print("No vertices")
            continue
        f.write("o Surface_%d\n" % o)
        #f.write("o %s\n" % obj.name)
        if m: f.write("mtllib " + os.path.splitext(os.path.basename(filepath))[0] + ".mtl\n")
        indices = data.indices.reshape((-1,3))
        normals = data.normals.reshape((-1,3))
        texcoords = data.texcoords.reshape((-1,2))
        #Calculate texcoords from colour values?
        if len(texcoords) == 0 and obj["texture"] == 'colourmap':
            label = obj["colourby"]
            if isinstance(label,int):
                #Use the given label index
                sets = list(datasets.keys())
                label = sets[label]
            elif len(label) == 0:
                #Use the default label
                label = 'values'
            valdata = data[label]
            if len(valdata) >= o+1:
                #Found matching value array
                v = valdata
                #Normalise [0,1]
                texcoords = (v - numpy.min(v)) / numpy.ptp(v)
                #Add 2nd dimension (not actually necessary,
                #tex coords can by 1d but breaks some loaders (meshlab)
                zeros = numpy.zeros((len(texcoords)))
                texcoords = numpy.vstack((texcoords,zeros)).reshape([2, -1]).transpose()

        #Colours?
        cs0 = ""
        vperc = 1
        if colourdict and len(data.colours):
            vperc = int(verts.shape[0] / len(data.colours))

        if verbose: print("- Writing vertices:",verts.shape)
        for v in verts:
            if colourdict:
                c = data.colours[ci]
                rgb = colour2rgb(c)
                f.write("v %.6f %.6f %.6f %.6f %.6f %.6f\n" % (v[0], v[1], v[2], rgb[0]/255.0, rgb[1]/255.0, rgb[2]/255.0))
            else:
                f.write("v %.6f %.6f %.6f\n" % (v[0], v[1], v[2]))
        if verbose: print("- Writing normals:",normals.shape)
        for n in normals:
            f.write("vn %.6f %.6f %.6f\n" % (n[0], n[1], n[2]))
        if verbose: print("- Writing texcoords:",texcoords.shape)
        if len(texcoords.shape) == 2:
            for t in texcoords:
                f.write("vt %.6f %.6f\n" % (t[0], t[1]))
        else:
            for t in texcoords:
                f.write("vt %.6f\n" % t)

        #Face elements v/t/n v/t v//n
        f.write(mtl_line)
        if len(normals) and len(texcoords):
            if verbose: print("- Writing faces (v/t/n):",indices.shape)
        elif len(texcoords):
            if verbose: print("- Writing faces (v/t):",indices.shape)
        elif len(normals):
            if verbose: print("- Writing faces (v//n):",indices.shape)
        else:
            if verbose: print("- Writing faces (v):",indices.shape)
        if verbose: print("- Colours :",data.colours.shape)
        if verbose: print("- Indices :",indices.shape)

        for n,i in enumerate(indices):
            if verbose and n%1000==0: print(".", end=''); sys.stdout.flush()
            i0 = i[0]+offset
            i1 = i[1]+offset
            i2 = i[2]+offset
            if colourdict:
                """
                ci = int(i[0] / vperc)
                #print(i0,vperc,ci)
                #print(data.colours)
                #c = data.colours[ci]
                c = data.colours[ci]
                rgb = colour2rgb(c)
                cs = colour2hex(rgb)
                cs1 = cs[1:]
                #if i[0]%100==0: print(ci,c,rgb,cs,mtl_line2)
                if cs0 != cs1 and cs in colourdict:
                    f.write("usemtl " + cs1 + "\n")
                    cs0 = cs1
                """

            if len(normals) and len(texcoords):
                f.write("f %d/%d/%d %d/%d/%d %d/%d/%d\n" % (i0, i0, i0, i1, i1, i1, i2, i2, i2))
            elif len(texcoords):
                f.write("f %d/%d %d/%d %d/%d\n" % (i0, i0, i1, i1, i2, i2))
            elif len(normals):
                f.write("f %d//%d %d//%d %d//%d\n" % (i0, i0, i1, i1, i2, i2))
            else:
                f.write("f %d %d %d\n" % (i0, i1, i2))
        if verbose: print()

        offset += verts.shape[0]
    return offset

def export_PLY(filepath, source, binary=True):
    """
    Export given object(s) to a PLY file
    Supports points or triangle mesh object data

    If source is lavavu.Viewer() exports all objects
    If source is lavavu.Object() exports single object

    Parameters
    ----------
    filepath : str
        Output file to write
    source : lavavu.Viewer or lavavu.Object
        Where to get object data to export
    binary : boolean
        Write vertex/face data as binary, default True
    """
    objects = _get_objects(source)
    with open(filepath, mode='wb') as f:
        voffset = 0
        foffset = 0
        #First count vertices, faces
        fc = 0
        vc = 0
        for obj in objects:
            for o,data in enumerate(obj):
                vc += len(data.vertices) #Vertices now in shape (N,3) so len returns correct count
                fc += len(data.indices) // 3

        #Pass the counts first
        vertex = None
        face = None
        print(vc, " vertices, ", fc, " faces")
        for obj in objects:
            for o,data in enumerate(obj):
                print("[%s] element %d of %d, type %s" % (obj.name, o+1, len(obj.data), data.type))
                #print("OFFSETS:",voffset,foffset)
                verts = data.vertices.reshape((-1,3))
                if len(verts) == 0:
                    print("No vertices")
                    return
                if len(verts) != vc:
                    print("Vertex count error!", len(verts), vc, verts.shape)
                    return
                indices = data.indices.reshape((-1,3))
                normals = data.normals.reshape((-1,3))
                texcoords = data.texcoords.reshape((-1,2))

                vperc = 0
                cperf = 0
                if len(data.colours):
                    vperc = int(verts.shape[0] / len(data.colours))
                    C = data.colours
                    #print("COLOURS:",len(C),C.shape, verts.shape[0], verts.shape[0] / len(data.colours), vperc)

                if data.type != 'points':
                    #Per face colours, or less
                    if face is None:
                        if vperc and vperc < len(verts):
                            #cperf = int(indices.shape[0] / len(data.colours))
                            face = numpy.zeros(shape=(fc), dtype=[('vertex_indices', 'i4', (3,)), ('red', 'u1'), ('green', 'u1'), ('blue', 'u1')])
                        else:
                            face = numpy.zeros(shape=(fc), dtype=[('vertex_indices', 'i4', (3,))])
                        print("FACE:",face.dtype)

                    for i,idx in enumerate(indices):
                        if i%1000==0: print(".", end=''); sys.stdout.flush()
                        if vperc and vperc < len(verts):
                            #Have colour, but less than vertices, apply to faces
                            ci = idx[0] // vperc
                            c = data.colours[ci]
                            rgb = colour2rgb(c)
                            face[i+foffset] = ([idx[0]+voffset, idx[1]+voffset, idx[2]+voffset], rgb[0], rgb[1], rgb[2])
                        else:
                            face[i+foffset] = ([idx[0]+voffset, idx[1]+voffset, idx[2]+voffset])
                    print()

                #Construct and write vertex elements
                if vertex is None:
                    #Setup vertex array based on first object element
                    D = [('x', 'f4'), ('y', 'f4'), ('z', 'f4')]
                    if normals.shape[0] == verts.shape[0]:
                        D += [('nx', 'f4'), ('ny', 'f4'), ('nz', 'f4')]
                    if texcoords.shape[0] == verts.shape[0]:
                        D += [('s', 'f4'), ('t', 'f4')]
                    if vperc and vperc == 1:
                        D += [('red', 'u1'), ('green', 'u1'), ('blue', 'u1')]
                    print("VERTEX:",D)
                    vertex = numpy.zeros(shape=(vc), dtype=D)

                for i,v in enumerate(verts):
                    #if i%1000==0:
                    #    print("vert index",i,vperc)
                    if i%1000==0: print(".", end=''); sys.stdout.flush()
                    E = [v[0], v[1], v[2]]
                    if normals.shape[0] == verts.shape[0]:
                        N = normals[i]
                        E += [N[0], N[1], N[2]]
                    if texcoords.shape[0] == verts.shape[0]:
                        T = texcoords[i]
                        E += [T[0], T[1]]
                    if vperc and vperc == 1:
                        c = data.colours[i]
                        rgb = colour2rgb(c)
                        E += [rgb[0], rgb[1], rgb[2]]
                    vertex[i+voffset] = tuple(E)
                print()

                #Update offsets : number of vertices / faces added
                voffset += verts.shape[0]
                foffset += indices.shape[0]

        import plyfile
        #vertex = numpy.array(vertex, dtype=vertex.dtype)
        els = []
        els.append(plyfile.PlyElement.describe(vertex, 'vertex'))
        if face is not None:
            els.append(plyfile.PlyElement.describe(face, 'face'))

        #Write, text or binary
        if binary:
            print("Writing binary PLY data")
            plyfile.PlyData(els).write(f)
        else:
            print("Writing ascii PLY data")
            plyfile.PlyData(els, text=True).write(f)

def _get_PLY_colours(element):
    """
    Extract colour data from PLY element and return as a numpy rgba array
    """
    r = None
    g = None
    b = None
    a = None
    #print(element.properties)
    for prop in element.properties:
        #print(prop,prop.name,prop.dtype)
        if 'red' in prop.name: r = element[prop.name]
        if 'green' in prop.name: g = element[prop.name]
        if 'blue' in prop.name: b = element[prop.name]
        if 'alpha' in prop.name: a = element[prop.name]

    if r is not None and g is not None and b is not None:
        if a is None:
            a = numpy.full(r.shape, 255)
        #return numpy.array([r, g, b, a])
        return numpy.vstack((r,g,b,a)).reshape([4, -1]).transpose()

    return None

def plot_PLY(lv, filename):
    """
    Plot triangles from a PlyData instance. Assumptions:
    has a 'vertex' element with 'x', 'y', and 'z' properties,
    has a 'face' element with an integral list property 'vertex_indices',
    all of whose elements have length 3.
    """
    import plyfile
    plydata = plyfile.PlyData.read(filename)

    x = plydata['vertex']['x']
    y = plydata['vertex']['y']
    z = plydata['vertex']['z']
    V = numpy.vstack((x,y,z)).reshape([3, -1]).transpose()
    #V = numpy.array([x, y, z])
    #print("VERTS:", V.shape)

    vp = []
    for prop in plydata['vertex'].properties:
        vp.append(prop.name)
        print(prop.name)
    
    N = None
    if 'nx' in vp and 'ny' in vp and 'nz' in vp:
        nx = plydata['vertex']['nx']
        ny = plydata['vertex']['ny']
        nz = plydata['vertex']['nz']
        N = numpy.vstack((nx,ny,nz)).reshape([3, -1]).transpose()
    
    T = None
    if 's' in vp and 't' in vp:
        s = plydata['vertex']['s']
        t = plydata['vertex']['t']
        T = numpy.vstack((s,t)).reshape([2, -1]).transpose()

    C = _get_PLY_colours(plydata['vertex'])

    if 'face' in plydata:
        #print(plydata['face'])
        #Face colours?
        if C is None:
            C = _get_PLY_colours(plydata['face'])

        tri_idx = plydata['face']['vertex_indices']
        idx_dtype = tri_idx[0].dtype
        tri_idx = numpy.asarray(tri_idx).flatten()
        #print(type(tri_idx),idx_dtype,tri_idx.shape)
        #print(tri_idx)

        triangles = numpy.array([t.tolist() for t in tri_idx]).flatten()
        #triangles = numpy.fromiter(tri_idx, [('data', idx_dtype, (3,))], count=len(tri_idx))['data']

        return lv.triangles(vertices=V, indices=triangles, colours=C, normals=N, texcoords=T)
    else:
        return lv.points(vertices=V, colours=C, normals=N, texcoords=T)

